#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Ruddr updater for Gandi LiveDNS v5 API"""

import ipaddress
from json import JSONDecodeError
from pprint import pprint
from typing import Optional, Tuple, List, Union, cast

import requests

from ..configuration import USER_AGENT
from ..exceptions import ConfigError, PublishError
from .updater import TwoWayZoneUpdater


class GandiUpdater(TwoWayZoneUpdater):
    """Ruddr updater for Gandi LiveDNS v5 API

    :param name: Name of the updater (from config section heading)
    :param addrfile: The :class:`~ruddr.Addrfile` object
    :param config: Dict of config options for this updater
    """

    def __init__(self, name, addrfile, config):
        super().__init__(name, addrfile, config['datadir'])

        # Validate FQDNs and store as list of (subdomain, domain) tuples
        try:
            fqdns = config['fqdns']
        except KeyError:
            self.log.critical("'fqdns' config option is required")
            raise ConfigError(f"{self.name} updater requires 'fqdns' config "
                              "option") from None
        self.init_hosts_and_zones(fqdns)

        # Gandi API key
        try:
            self.api_key = config['api_key']
        except KeyError:
            self.log.critical("'api_key' config option is required")
            raise ConfigError(f"{self.name} updater requires 'api_key' config "
                              "option") from None

        # Gandi API endpoint - base URL to use for the LiveDNS API. Normally
        # not required, but can be used if you wish to test in the sandbox
        # API environment
        self.endpoint = config.get('endpoint',
                                   'https://api.gandi.net/v5/livedns')

    def _api_request(self, method, api, params=None, data=None):
        """Issue a LiveDNS API request.

        :param method: HTTP method, ``'GET'`` or ``'PUT'``
        :param api: Specific API to access, e.g. ``'/dns/rrtypes'``
        :param params: A dict of URL parameters (i.e. the key=values that go
                       after the question mark in the URL)
        :param data: A JSON-serializable dict to become the request body.

        :return: The :class:`Response` object, or `None` if there was an error
                 (which will be logged)
        """
        headers = {'Authorization': "Apikey " + self.api_key,
                   'User-Agent': USER_AGENT}
        if method == 'GET':
            method_f = requests.get
        elif method == 'PUT':
            method_f = requests.put
        url = self.endpoint + api
        try:
            if params is None and data is None:
                r = method_f(url, headers=headers)
            elif params is None:
                r = method_f(url, headers=headers, json=data)
            elif data is None:
                r = method_f(url, headers=headers, params=params)
            else:
                r = method_f(url, headers=headers, params=params, json=data)
        except requests.exceptions.RequestException as e:
            self.log.error("Could not %s %s: %s", method, url, e)
            return None
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError:
            self.log.error("Received HTTP %d when trying to %s %s:\n%s",
                           r.status_code, method, url, r.text)
            return None

        try:
            obj = r.json()
        except JSONDecodeError:
            self.log.error("Could not parse JSON response from %s %s:\n%s",
                           method, url, r.text)
            return None
        return obj

    def get_zones(self):
        response = self._api_request('GET', '/domains')
        if response is None:
            raise PublishError("Could not fetch zones for updater "
                               f"{self.name}")
        try:
            result = [rec['fqdn'] for rec in response]
        except (KeyError, TypeError):
            self.log.error("Unknown response structure from /domains:\n%s",
                           pprint(response))
            raise PublishError("Unknown response structure from /domains")
        return result

    def _fetch_zone_records(self, zone: str, rec_type: str) -> Union[
        List[Tuple[str, ipaddress.IPv4Address, int]],
        List[Tuple[str, ipaddress.IPv6Address, int]],
    ]:
        """Fetch A or AAAA records for the given zone

        :param zone: The zone to fetch records for
        :param rec_type: ``'A'`` or ``'AAAA'``
        :return: A list of 3-tuples ``(subdomain, addr, ttl)`` as returned by
                 :meth:`fetch_zone_ipv4s` and :meth:`fetch_zone_ipv6s`
        """
        assert rec_type in ('A', 'AAAA')
        api = f'/domains/{zone}/records'
        params = {'rrset_type': rec_type}
        response = self._api_request('GET', api, params)
        if response is None:
            raise PublishError(f"Could not fetch {rec_type} records for zone "
                               f"'{zone}'")

        result: Union[
            List[Tuple[str, ipaddress.IPv4Address, int]],
            List[Tuple[str, ipaddress.IPv6Address, int]],
        ] = []

        try:
            for rec in response:
                name = rec['rrset_name']
                if name == '@':
                    name = ''
                ttl = rec['rrset_ttl']
                for ip in rec['rrset_values']:
                    if rec_type == 'A':
                        ip = ipaddress.IPv4Address(ip)
                    else:
                        ip = ipaddress.IPv6Address(ip)
                    result.append((name, ip, ttl))
        except ipaddress.AddressValueError:
            self.log.error("Invalid IP from %s:\n%s", api, pprint(response))
            raise PublishError(f"Invalid IP from {api}")
        except (KeyError, TypeError):
            self.log.error("Unknown response structure from %s:\n%s",
                           api, pprint(response))
            raise PublishError(f"Unknown response structure from {api}")
        return result

    def fetch_zone_ipv4s(
        self,
        zone: str
    ) -> List[Tuple[str, ipaddress.IPv4Address, Optional[int]]]:
        return cast(List[Tuple[str, ipaddress.IPv4Address, int]],
                    self._fetch_zone_records(zone, 'A'))

    def fetch_zone_ipv6s(
        self,
        zone: str
    ) -> List[Tuple[str, ipaddress.IPv6Address, Optional[int]]]:
        return cast(List[Tuple[str, ipaddress.IPv6Address, int]],
                    self._fetch_zone_records(zone, 'AAAA'))

    def _put_record(self, zone: str, subdomain: str,
                    rec_type: str, addrs: List[str], ttl: int):
        if subdomain == '':
            subdomain = '@'
        api = f'/domains/{zone}/records/{subdomain}/{rec_type}'
        data = {'rrset_values': addrs, 'rrset_ttl': ttl}
        response = self._api_request('PUT', api, data=data)
        if response is None:
            raise PublishError(f"Could not PUT {api}")

    def put_subdomain_ipv4(self, subdomain: str, zone: str,
                           address: ipaddress.IPv4Address, ttl: Optional[int]):
        assert ttl is not None
        addrs = [address.exploded]
        self._put_record(zone, subdomain, 'A', addrs, ttl)

    def put_subdomain_ipv6s(self, subdomain: str, zone: str,
                            addresses: List[ipaddress.IPv6Address],
                            ttl: Optional[int]):
        assert ttl is not None
        addrs = [address.compressed for address in addresses]
        self._put_record(zone, subdomain, 'AAAA', addrs, ttl)



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pytest

import doubles


def test_setup_teardown_not_implemented(advance):
    """Test starting, initial check, and teardown works when setup and teardown
    not implemented"""
    notifier = doubles.MockNotifier(
        'mock_notifier',
        dict(),
        setup_implemented=False,
        teardown_implemented=False,
    )
    notifier.set_check_intervals(retry_min_interval=1,
                                 retry_max_interval=5)

    assert notifier.call_sequence == []
    notifier.start()
    notifier.join_first_check()
    assert notifier.call_sequence == ['setup', 'check']
    advance.by(10)

    assert advance.count_running() == 0, "Unexpected checks pending"
    assert notifier.call_sequence == ['setup', 'check']
    notifier.stop()
    assert advance.count_running() == 0, "Unexpected pending checks after stop"
    assert notifier.call_sequence == ['setup', 'check', 'teardown']


def test_check_not_implemented(advance):
    """Test starting and teardown works when check_once not implemented"""
    notifier = doubles.MockNotifier(
        'mock_notifier',
        dict(),
        check_implemented=False,
    )
    notifier.set_check_intervals(retry_min_interval=1,
                                 retry_max_interval=5)

    assert notifier.call_sequence == []
    notifier.start()
    notifier.join_first_check()
    assert notifier.call_sequence == ['setup', 'check']
    advance.by(10)

    assert advance.count_running() == 0, "Unexpected checks pending"
    assert notifier.call_sequence == ['setup', 'check']
    notifier.stop()
    assert advance.count_running() == 0, "Unexpected pending checks after stop"
    assert notifier.call_sequence == ['setup', 'check', 'teardown']


def test_setup_teardown_not_implemented_with_scheduled_checks(advance):
    """Test starting, initial check, and teardown works when setup and teardown
    not implemented"""
    notifier = doubles.MockNotifier(
        'mock_notifier',
        dict(),
        setup_implemented=False,
        teardown_implemented=False,
    )
    notifier.set_check_intervals(retry_min_interval=1,
                                 retry_max_interval=5,
                                 success_interval=7)

    assert notifier.call_sequence == []
    notifier.start()
    notifier.join_first_check()
    assert notifier.call_sequence == ['setup', 'check']
    advance.by(6)

    notifier.do_notify()
    assert notifier.check_count == 2
    advance.by(7)
    assert notifier.check_count == 3
    advance.by(6)

    assert notifier.call_sequence == ['setup',
                                      'check', 'check', 'check']
    notifier.stop()
    assert advance.count_running() == 0, "Unexpected pending checks after stop"
    assert notifier.call_sequence == ['setup',
                                      'check', 'check', 'check',
                                      'teardown']


def test_default_check_intervals(advance):
    """Test that the default retry and success intervals are as expected"""
    notifier = doubles.MockNotifier(
        'mock_notifier',
        dict(),
        [False, False, False, False, False, False, False, False, False, False,
         True],
    )

    assert notifier.call_sequence == []
    notifier.start()
    notifier.join_first_check()
    assert notifier.call_sequence == ['setup', 'check']
    advance.by(300)

    assert notifier.check_count == 2
    advance.by(600)
    assert notifier.check_count == 3
    advance.by(1200)
    assert notifier.check_count == 4
    advance.by(2400)
    assert notifier.check_count == 5
    advance.by(4800)
    assert notifier.check_count == 6
    advance.by(9600)
    assert notifier.check_count == 7
    advance.by(19200)
    assert notifier.check_count == 8
    advance.by(38400)
    assert notifier.check_count == 9
    advance.by(76800)
    assert notifier.check_count == 10
    advance.by(86400)
    assert notifier.check_count == 11
    advance.by(100000)

    assert advance.count_running() == 0, "Unexpected checks pending"
    assert notifier.teardown_count == 0
    assert notifier.check_count == 11
    notifier.stop()
    assert advance.count_running() == 0, "Unexpected pending checks after stop"
    assert notifier.setup_count == 1
    assert notifier.check_count == 11
    assert notifier.teardown_count == 1


def test_set_intervals_from_config(advance):
    """Test that the retry and success intervals can be overridden by config"""
    notifier = doubles.MockNotifier(
        'mock_notifier',
        dict(),
        [False, False, False, False, True, True],
    )
    notifier.set_check_intervals(retry_min_interval=300,
                                 retry_max_interval=86400,
                                 success_interval=0,
                                 config={
                                     'retry_min_interval': '1',
                                     'retry_max_interval': '5',
                                     'interval': '7',
                                 })

    assert notifier.call_sequence == []
    notifier.start()
    notifier.join_first_check()
    assert notifier.call_sequence == ['setup', 'check']
    advance.by(1)

    assert notifier.check_count == 2
    advance.by(2)
    assert notifier.check_count == 3
    advance.by(4)
    assert notifier.check_count == 4
    advance.by(5)
    assert notifier.check_count == 5
    advance.by(10)

    assert advance.count_running() == 0, "Unexpected checks pending"
    assert notifier.teardown_count == 0
    assert notifier.check_count == 5
    notifier.stop()
    assert advance.count_running() == 0, "Unexpected pending checks after stop"
    assert notifier.setup_count == 1
    assert notifier.check_count == 5
    assert notifier.teardown_count == 1


def test_config_not_override_zero_success_interval(advance):
    """Test that a zero success interval (no repeat) isn't overridden by
    config"""
    notifier = doubles.MockNotifier(
        'mock_notifier',
        dict(),
    )
    notifier.set_check_intervals(retry_min_interval=300,
                                 retry_max_interval=86400,
                                 success_interval=0,
                                 config={
                                     'retry_min_interval': '1',
                                     'retry_max_interval': '5',
                                     'interval': '7',
                                 })

    assert notifier.call_sequence == []
    notifier.start()
    notifier.join_first_check()
    assert notifier.call_sequence == ['setup', 'check']
    advance.by(100000)

    assert advance.count_running() == 0, "Unexpected checks pending"
    assert notifier.teardown_count == 0
    assert notifier.check_count == 1
    notifier.stop()
    assert advance.count_running() == 0, "Unexpected pending checks after stop"
    assert notifier.setup_count == 1
    assert notifier.check_count == 1
    assert notifier.teardown_count == 1


# Sequence is list of checks as tuples (successful, on_demand, interval_after)
# on_demand is ignored for the initial check
@pytest.mark.parametrize("sequence,polling", [
    # === Non-polling tests ===
    # setup, initial check, teardown
    ([
        (True, False, 10),
     ], False),
    # setup, initial check, on demand check, teardown
    ([
         (True, False, 10),
         (True, True, 10),
     ], False),
    # setup, initial check fails, retry succeeds, teardown
    ([
         (False, False, 1),
         (True, False, 10),
     ], False),
    # setup, initial check fails, on demand check before retry, retry doesn't
    # happen, teardown
    ([
         (False, False, 0.5),
         (True, True, 10),
     ], False),
    # setup, initial check fails, next check fails, next check passes, teardown
    ([
         (False, False, 1),
         (False, False, 2),
         (True, False, 10),
     ], False),
    # setup, initial check fails, next check fails, on demand check fails but
    # short timeout again, then teardown
    ([
         (False, False, 1),
         (False, False, 1.5),
         (False, True, 1),
         (True, False, 10),
     ], False),

    # === Polling tests ===
    # setup, initial check, on demand check, scheduled check, teardown, no more
    # check
    ([
         (True, False, 6),
         (True, True, 7),
         (True, False, 6),
     ], True),
    # setup, initial check, scheduled check, teardown, no more check
    ([
         (True, False, 7),
         (True, False, 6),
     ], True),
    # setup, initial check fails, retry succeeds, scheduled check, teardown
    ([
         (False, False, 1),
         (True, False, 7),
         (True, False, 6),
     ], True),
    # setup, initial check, scheduled check fails, retry succeeds, scheduled
    # check, teardown
    ([
         (True, False, 7),
         (False, False, 1),
         (True, False, 7),
         (True, False, 6),
     ], True),
    # setup, initial check fails, retry succeeds, on demand check, scheduled
    # check, teardown
    ([
         (False, False, 1),
         (True, False, 6),
         (True, True, 7),
         (True, False, 6),
     ], True),
    # setup, initial check fails, retry fails 3 times, retry succeeds,
    # scheduled check, teardown
    ([
         (False, False, 1),
         (False, False, 2),
         (False, False, 4),
         (False, False, 5),
         (True, False, 7),
         (True, False, 6),
     ], True),
    # setup, initial check fails, retry fails 4 times, teardown, no more check
    ([
         (False, False, 1),
         (False, False, 2),
         (False, False, 4),
         (False, False, 5),
         (False, False, 4),
     ], True),
    # setup, initial check fails, fails twice more, on demand check fails,
    # retry after minimum delay fails, retry succeeds, teardown
    ([
         (False, False, 1),
         (False, False, 2),
         (False, False, 3),
         (False, True, 1),
         (False, False, 2),
         (True, False, 6),
     ], True),
])
def test_notifier_sequence(sequence, polling, advance):
    notifier = doubles.MockNotifier(
        'mock_notifier',
        dict(),
        [c[0] for c in sequence],
    )
    if polling:
        notifier.set_check_intervals(retry_min_interval=1,
                                     retry_max_interval=5,
                                     success_interval=7)
    else:
        notifier.set_check_intervals(retry_min_interval=1,
                                     retry_max_interval=5)

    # Setup and first check
    assert notifier.call_sequence == []
    notifier.start()
    notifier.join_first_check()
    assert notifier.call_sequence == ['setup', 'check']
    advance.by(sequence[0][2])

    # Rest of the checks
    check_count = 1
    for _, on_demand, interval_after in sequence[1:]:
        if on_demand:
            notifier.do_notify()
        check_count += 1
        assert notifier.check_count == check_count
        advance.by(interval_after)

    # If not polling, no checks pending (e.g. no extra retry)
    if not polling:
        assert advance.count_running() == 0, "Unexpected checks pending"

    # Teardown and final validations
    assert notifier.teardown_count == 0
    notifier.stop()
    assert advance.count_running() == 0, "Unexpected pending checks after stop"
    assert notifier.setup_count == 1
    assert notifier.check_count == len(sequence)
    assert notifier.teardown_count == 1



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""All Ruddr exceptions"""


class RuddrException(Exception):
    """Base class for all Ruddr exceptions"""


class RuddrSetupError(RuddrException):
    """Base class for Ruddr exceptions that happen during startup"""


class ConfigError(RuddrSetupError):
    """Raised when the configuration is malformed or has other errors"""


class NotifierSetupError(RuddrSetupError):
    """Raised by a notifier when there is a fatal error during setup for
    persistent checks"""


class NotStartedError(RuddrException):
    """Raised when requesting an on-demand notify before starting the
    notifier"""


class NotifyError(RuddrException):
    """Notifiers should raise when an attempt to check the current IP address
    fails. Doing so triggers the retry mechanism."""


class PublishError(RuddrException):
    """Updaters should raise when an attempt to publish an update fails. Doing
    so triggers the retry mechanism."""


class FatalPublishError(PublishError):
    """Updaters should raise when an attempt to publish an update fails in a
    non-recoverable way. Doing so causes the updater to halt until the next
    time Ruddr is started."""



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Ruddr updater for providers using the de facto standard /nic/update API"""

import enum
import ipaddress
from typing import Dict

import requests

from ..configuration import USER_AGENT
from ..exceptions import ConfigError, PublishError, FatalPublishError
from .updater import OneWayUpdater


class _IPv6Dialect(enum.Enum):
    SEPARATE = "separate"
    SEPARATE_NO = "separate_no"
    COMBINED = "combined"


class StandardUpdater(OneWayUpdater):
    """Ruddr updater for providers using the de facto standard /nic/update API

    :param name: Name of the updater (from config section heading)
    :param addrfile: The :class:`~ruddr.Addrfile` object
    :param config: Dict of config options for this updater
    """

    def __init__(self, name, addrfile, config: dict):
        super().__init__(name, addrfile)

        # Minimum retry interval
        try:
            min_retry = int(config.get('min_retry_interval', '300'))
        except ValueError:
            self.log.critical("'min_retry_interval' config option must be an "
                              "integer")
            raise ConfigError(f"{name} updater requires an integer for "
                              "'min_retry_interval' config option")

        # Hosts
        try:
            hosts = config['hosts']
        except KeyError:
            self.log.critical("'hosts' config option is required")
            raise ConfigError(f"{self.name} updater requires 'hosts' config "
                              "option") from None

        # Nameserver
        nameserver = config.get('nameserver')

        self.init_params(hosts, nameserver, min_retry)

        # Endpoint
        try:
            self.endpoint = config['endpoint']
        except KeyError:
            self.log.critical("'endpoint' config option is required")
            raise ConfigError(f"{self.name} updater requires 'endpoint' config"
                              " option") from None
        if self.endpoint[-1] == '/':
            self.endpoint = self.endpoint[:-1]
        self.endpoint += "/nic/update"

        # Username
        try:
            username = config['username']
        except KeyError:
            self.log.critical("'username' config option is required")
            raise ConfigError(f"{self.name} updater requires 'username' config"
                              " option") from None

        # Password
        try:
            password = config['password']
        except KeyError:
            self.log.critical("'password' config option is required")
            raise ConfigError(f"{self.name} updater requires 'password' config"
                              " option") from None

        self.auth = (username, password)

        # IPv6 dialect
        ipv6_dialect = config.get('ipv6_dialect', 'separate')
        try:
            self.ipv6_dialect = _IPv6Dialect(ipv6_dialect)
        except ValueError:
            self.log.critical(f"IPv6 dialect '{ipv6_dialect}' is not a valid "
                              "choice")
            raise ConfigError(f"{self.name} updater has invalid IPv6 dialect "
                              f"{ipv6_dialect}") from None

    def publish_ipv4_one_host(self,
                              hostname: str,
                              address: ipaddress.IPv4Address):
        """Attempt to publish an IPv4 address for a single host

        :param hostname: The host to publish for
        :param address: The address to publish

        :raise PublishError: if publishing fails
        :raise FatalPublishError: if publishing fails in a non-recoverable way
                                  (all future publishing will halt)
        """
        params = {'hostname': hostname}
        if self.ipv6_dialect is _IPv6Dialect.SEPARATE_NO:
            params['myip'] = address.exploded
            params['myipv6'] = 'no'
        else:
            params['myip'] = address.exploded

        self._send_request(params, hostname, 'IPv4', address.exploded)

    def publish_ipv6_one_host(self,
                              hostname: str,
                              address: ipaddress.IPv6Address):
        """Attempt to publish an IPv6 address for a single host

        :param hostname: The host to publish for
        :param address: The address to publish

        :raise PublishError: if publishing fails
        :raise FatalPublishError: if publishing fails in a non-recoverable way
                                  (all future publishing will halt)
        """
        params = {'hostname': hostname}
        if self.ipv6_dialect is _IPv6Dialect.SEPARATE_NO:
            params['myip'] = 'no'
            params['myipv6'] = address.compressed
        elif self.ipv6_dialect is _IPv6Dialect.COMBINED:
            params['myip'] = address.compressed
        else:
            params['myipv6'] = address.compressed

        self._send_request(params, hostname, 'IPv6', address.compressed)

    def _send_request(self,
                      params: Dict[str, str],
                      hostname: str,
                      addr_type: str,
                      addr: str):
        """Send an update request and handle the reply

        :param params: Dict of parameters to send with the request
        :param hostname: Name of the host being updated
        :param addr_type: ``'IPv4'`` or ``'IPv6'``
        :param addr: The address being assigned to the host

        :raises PublishError: if there is an error
        :raises FatalPublishError: if there is a non-retryable error
        """
        try:
            r = requests.get(self.endpoint,
                             auth=self.auth,
                             params=params,
                             headers={'User-Agent': USER_AGENT})
        except requests.exceptions.RequestException as e:
            self.log.error("Could not update hostname '%s' %s to %s: %s",
                           hostname, addr_type, addr, e)
            raise PublishError("Updater %s could not access %s: %s" % (
                self.name, self.endpoint, e)) from e

        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            self.log.error("Received HTTP %d when trying to update hostname "
                           "'%s' %s to %s:\n%s", r.status_code,
                           hostname, addr_type, addr, r.text)
            raise PublishError("Updater %s got HTTP %d for %s" % (
                self.name, r.status_code, self.endpoint)) from e

        response = r.text.strip().split()
        if response[0] == 'good':
            self.log.info("Hostname '%s' %s updated to %s",
                          hostname, addr_type, addr)
            return
        if response[0] == 'nochg':
            self.log.info("Hostname '%s' %s already set to %s",
                          hostname, addr_type, addr)
            return
        # Different providers have different error codes. The following are for
        # server-side issues at various providers, indicating a retry is
        # reasonable.
        if response[0] in ('911', 'dnserr', 'servererror'):
            self.log.error("Server returned response '%s' when trying to "
                           "update hostname '%s' %s to %s",
                           r.text, hostname, addr_type, addr)
            raise PublishError("Updater %s got response from server: %s" %
                               (self.name, r.text))
        # Other return codes are likely client-side issues (e.g. bad config,
        # agent is blocked). Do not retry.
        else:
            self.log.error("Server returned response '%s' when trying to "
                           "update hostname '%s' %s to %s",
                           r.text, hostname, addr_type, addr)
            raise FatalPublishError("Updater %s got response from server: %s" %
                                    (self.name, r.text))



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Ruddr configuration parsing"""

import configparser
import os.path
import pathlib
import sys

from typing import Dict, Callable, Union, Optional, Iterable

if sys.version_info < (3, 10):
    from importlib_metadata import version
else:
    from importlib.metadata import version

from .exceptions import ConfigError


USER_AGENT = f"ruddr/{version('ruddr')} (ruddr@dcpx.org)"


DEFAULT_DATA_DIR = '/var/lib/ruddr'


class Config:
    """Contains all Ruddr configuration required by
    :class:`~ruddr.DDNSManager`. Normally, this would be created from a
    configuration file by :func:`read_config` or :func:`read_config_from_path`,
    but it can also be created directly when using Ruddr as a library.

    Note that all configuration values should be strings, as they would be from
    Python's :class:`configparser.ConfigParser`.

    The configuration must be finalized before use, but programs using Ruddr as
    a library need not concern themselves with that.
    :class:`~ruddr.DDNSManager` will do that itself.

    :param main: A dictionary of global configuration options, that is, the
                 options that go under ``[ruddr]`` in the configuration file.
    :param notifiers: A dictionary of notifier configurations. Keys are
                      notifier names (i.e. the ``XYZ`` part of
                      ``[notifier.XYZ]`` if it were from a configuration file)
                      and values are themselves dicts of config options for
                      that notifier.
    :param updaters: A dictionary of updater configurations. Keys are updater
                     names (i.e. the ``XYZ`` part of ``[updater.XYZ]`` if it
                     were from a configuration file) and values are themselves
                     dicts of config options for that updater.
    """

    def __init__(self,
                 main: Dict[str, str],
                 notifiers: Dict[str, Dict[str, str]],
                 updaters: Dict[str, Dict[str, str]]):
        #: Dict containing global configuration (from the ``[ruddr]`` section)
        self._main: Dict[str, str] = main

        #: Notifier configurations (from ``[notifier.<name>]`` sections)
        self._notifiers: Dict[str, Dict[str, str]] = notifiers

        #: Updater configurations (from ``[updater.<name>]`` sections)
        self._updaters: Dict[str, Dict[str, str]] = updaters

        #: Whether the config has been finalized yet
        self._finalized = False

    def _check_finalized(self):
        """Raise an exception if the config is not finalized"""
        if not self._finalized:
            raise ConfigError("Tried to access config before it was finalized")

    @property
    def main(self):
        self._check_finalized()
        return self._main

    @property
    def notifiers(self):
        self._check_finalized()
        return self._notifiers

    @property
    def updaters(self):
        self._check_finalized()
        return self._updaters

    def _fill_defaults(self):
        """Fill in defaults if they are not yet set"""
        if 'datadir' not in self._main:
            self._main['datadir'] = DEFAULT_DATA_DIR
        if not os.path.isabs(self._main['datadir']):
            raise ConfigError("Config option 'datadir' cannot be a relative"
                              "path")

    def _validate_types(
        self,
        kind: str,
        config_dict: Dict[str, Dict[str, str]],
        validate_type: Callable[[Optional[str], str], bool],
    ) -> None:
        """Verify that updater or notifier types are assigned, that they are
        valid, and that there is at least one

        :param kind: ``'Updater'`` or ``'Notifier'``
        :param config_dict: The config dict for those items
        :param validate_type: A callable that validates the updater or notifier
                              names

        :raises ConfigError: if any type is missing or invalid
        """
        if len(config_dict) == 0:
            raise ConfigError(f"At least one {kind} is required")

        for name, config in config_dict.items():
            module = config.get('module')
            try:
                type_ = config['type']
            except KeyError:
                raise ConfigError(f"{kind} {name} requires a type") from None
            exists = validate_type(module, type_)
            if not exists and module is None:
                raise ConfigError(f"No built-in {kind} of type {type_}")
            elif not exists:
                raise ConfigError(f"{kind} module or class {module}.{type_} "
                                  "does not exist")

    def _validate_and_assign_notifiers(self) -> None:
        """Translate ``notifier`` keys on updaters to ``notifier4`` and
        ``notifier6`` and copy global notifiers into updater config if
        necessary. In the process, validate that all ``notifier{4|6}`` keys
        point to defined notifiers.

        :raises ConfigError: if the config is invalid
        """
        # Get default notifiers, ensure they exist, and error if "notifier" is
        #   set when "notifier4" and "notifier6" are both set
        if ('notifier' in self._main and
                'notifier4' in self._main and
                'notifier6' in self._main):
            raise ConfigError('Main Ruddr config does not need "notifier" when'
                              ' "notifier4" and "notifier6" are both set')
        try:
            g_notifier4 = self._main['notifier4']
        except KeyError:
            g_notifier4 = self._main.get('notifier')
        if g_notifier4 is not None and g_notifier4 not in self._notifiers:
            raise ConfigError("Notifier %s does not exist" % g_notifier4)
        try:
            g_notifier6 = self._main['notifier6']
        except KeyError:
            g_notifier6 = self._main.get('notifier')
        if g_notifier6 is not None and g_notifier6 not in self._notifiers:
            raise ConfigError("Notifier %s does not exist" % g_notifier6)

        for updater_name, updater_config in self._updaters.items():
            # Error if "notifier" is set when "notifier4" and "notifier6" are
            #   both also set
            if ('notifier' in updater_config and
                    'notifier4' in updater_config and
                    'notifier6' in updater_config):
                raise ConfigError('Updater %s does not need "notifier" when '
                                  '"notifier4" and "notifier6" are both set'
                                  % updater_name)

            try:
                notifier4 = updater_config['notifier4']
            except KeyError:
                notifier4 = updater_config.get('notifier')
            if notifier4 is not None and notifier4 not in self._notifiers:
                raise ConfigError("Notifier %s does not exist" % notifier4)

            try:
                notifier6 = updater_config['notifier6']
            except KeyError:
                notifier6 = updater_config.get('notifier')
            if notifier6 is not None and notifier6 not in self._notifiers:
                raise ConfigError("Notifier %s does not exist" % notifier6)

            # If no notifiers configured for updater, use default notifiers
            if notifier4 is None and notifier6 is None:
                if g_notifier4 is None and g_notifier6 is None:
                    raise ConfigError("No notifier is configured for updater "
                                      "%s and there are no default notifiers "
                                      "configured" % updater_name)
                else:
                    notifier4 = g_notifier4
                    notifier6 = g_notifier6

            try:
                del updater_config['notifier']
            except KeyError:
                pass
            # Both cannot be None, but one can
            if notifier4 is not None:
                updater_config['notifier4'] = notifier4
            if notifier6 is not None:
                updater_config['notifier6'] = notifier6

    def _copy_globals(self) -> None:
        """Copy relevant global (e.g. ``[ruddr]``) config options into updater
        and notifier configs. For example, the data directory.

        Does not copy the global notifier(s), which are handled by
        :func:`_validate_and_assign_notifiers`.
        """
        for notifier_config in self._notifiers.values():
            notifier_config['datadir'] = self._main['datadir']
        for updater_config in self._updaters.values():
            updater_config['datadir'] = self._main['datadir']

    def finalize(self,
                 validate_notifier_type: Callable[[Optional[str], str], bool],
                 validate_updater_type: Callable[[Optional[str], str], bool]):
        """Used by :class:`~ruddr.DDNSManager` to finalize the configuration.
        This consists of validating the updater and notifier types, filling
        default values, and doing some normalization.

        Programs using Ruddr as a library need not call this function
        themselves; :class:`~ruddr.DDNSManager` will handle it.

        :param validate_notifier_type: A callable to check if a notifier type
                                       is valid. First parameter is a module
                                       name, or ``None`` if it's a built-in
                                       notifier type. Second parameter is a
                                       class name or built-in notifier type
                                       name.
        :param validate_updater_type: A callable to check if an updater type is
                                      valid. Parameters are the same as above.
        :raises ConfigError: if the configuration is invalid
        """
        if self._finalized:
            return

        self._fill_defaults()
        self._validate_types('Updater', self._updaters, validate_updater_type)
        self._validate_types('Notifier', self._notifiers,
                             validate_notifier_type)
        self._validate_and_assign_notifiers()
        self._copy_globals()

        self._finalized = True


def _process_config(config: configparser.ConfigParser) -> Config:
    """Process the given :class:`~configparser.ConfigParser` into a
    :class:`Config`

    :param config: The configuration to process
    :raises ConfigError: if the configuration is invalid
    :returns: the processed and validated configuration
    """
    # Note: ConfigParser already handles catching duplicate sections and
    #   duplicate keys

    main: Dict[str, str] = dict()
    notifiers: Dict[str, Dict[str, str]] = dict()
    updaters: Dict[str, Dict[str, str]] = dict()

    for section in config.sections():
        if section == 'ruddr':
            main.update(config[section])
            continue

        kind, sep, name = section.partition('.')
        if kind == 'notifier':
            notifiers[name] = dict(config[section])
        elif kind == 'updater':
            updaters[name] = dict(config[section])
        else:
            raise ConfigError("Config section %s is not a notifier "
                              "or updater section" % section)
        if sep != '.' or name == '':
            raise ConfigError("Config section %s must have a '.<name>'" %
                              section)

    return Config(main, notifiers, updaters)


def read_config_from_path(filename: Union[str, pathlib.Path]) -> Config:
    """Read configuration from the named file or :class:`~pathlib.Path`

    :param filename: Filename or path to read from
    :raises ConfigError: if the config file cannot be read or is invalid
    :return: A :class:`Config` ready to be passed to
             :class:`~ruddr.DDNSManager`
    """
    try:
        with open(filename, 'r') as f:
            return read_config(f)
    except OSError as e:
        raise ConfigError("Could not read config file %s: %s" %
                          (filename, e.strerror)) from e


def read_config(configfile: Iterable[str]) -> Config:
    """Read configuration in from the given file-like object opened in text
    mode

    :param configfile: Filelike object to read the config from
    :raises ConfigError: if the config file cannot be read or is invalid
    :return: A :class:`Config` ready to be passed to
             :class:`~ruddr.DDNSManager`
    """
    config = configparser.ConfigParser()
    try:
        config.read_file(configfile)
    except configparser.Error as e:
        raise ConfigError("Error in config file: %s" % e) from e
    except OSError as e:
        raise ConfigError("Could not read config file: %s" % e) from e

    return _process_config(config)



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Ruddr notifier that checks the IP address using a what-is-my-ip-style
website"""

import ipaddress
import socket
import requests

from ruddr.configuration import USER_AGENT
from ruddr.exceptions import NotifyError, ConfigError
from ruddr.util import RequestsFamilyRestriction
from .notifier import Notifier


class WebNotifier(Notifier):
    """Ruddr notifier that checks the IP address using a what-is-my-ip-style
    website"""

    def __init__(self, name, config):
        super().__init__(name, config)

        # URL to request IP addresses from. Normally, both IPv4 and IPv6
        # addresses will be requested from the same URL (by issuing separate
        # requests over IPv4 and IPv6). If a different URL should be used for
        # IPv4 and IPv6, specify the IPv6 URL with "url6=".
        #
        # If only IPv4 is needed, use "notifier4=" to attach this notifier to
        # updaters in the config instead of plain "notifier=". When there are
        # no updaters requesting IPv6 addresses from this notifier, it will
        # skip checking the IPv6 address. (Vice versa for skipping IPv4.)
        try:
            self.url4 = config['url']
        except KeyError:
            self.log.critical("'url' config option is required. Only need "
                              "IPv6? Use 'url=' and attach this notifier to "
                              "an updater with 'notifier6='.")
            raise ConfigError(f"{self.name} notifier requires 'url' config "
                              "option") from None

        # URL to request IPv6 addresses from, if different from the URL to
        # request IPv4 addresses. (If only IPv6 is needed, there is no need to
        # use this option. Instead, use the regular 'url=' option, and attach
        # this notifier to an updater with 'notifier6='. It will only request
        # IPv6 addresses.)
        try:
            self.url6 = config['url6']
        except KeyError:
            self.url6 = self.url4

        # Timeout to use waiting for a response from the HTTP server, in
        # seconds
        try:
            self.timeout4 = float(config.get('timeout', '10'))
        except ValueError:
            self.log.critical("'timeout' config option must be a number")
            raise ConfigError(f"'timeout' option for {self.name} notifier "
                              "must be a number") from None

        # Timeout to use waiting for a response from the HTTP server, in
        # seconds
        try:
            self.timeout6 = float(config['timeout6'])
        except KeyError:
            self.timeout6 = self.timeout4
        except ValueError:
            self.log.critical("'timeout6' config option must be a number")
            raise ConfigError(f"'timeout6' option for {self.name} notifier "
                              "must be a number") from None

        # IPv6 prefix: Number of bits in the network prefix. Defaults to 64,
        # but can be manually specified in case your ISP delegates a shorter
        # prefix.
        try:
            self.ipv6_prefix = int(config.get('ipv6_prefix', '64'))
        except ValueError:
            self.log.critical("'ipv6_prefix' config option must be an integer "
                              "from 1-128")
            raise ConfigError(f"'ipv6_prefix' option for {self.name} notifier "
                              "must be an integer from 1-128") from None
        if not (1 <= self.ipv6_prefix <= 128):
            self.log.critical("'ipv6_prefix' config option must be an integer "
                              "from 1-128")
            raise ConfigError(f"'ipv6_prefix' option for {self.name} notifier "
                              "must be an integer from 1-128")

        self.set_check_intervals(retry_min_interval=60,
                                 retry_max_interval=86400,
                                 success_interval=10800,
                                 config=config)

    def check_once(self):
        self.log.info("Checking IP addresses.")

        # None if not wanted, otherwise True if assigned, False if not assigned
        got_ipv4 = None
        got_ipv6 = None
        # True if there is an HTTP error, suggesting something went wrong
        # rather than this host simply not having connectivity on that address
        # family right now (the latter case not always being a problem--but
        # we'll check later)
        err_ipv4 = False
        err_ipv6 = False

        if self.want_ipv4():
            with RequestsFamilyRestriction(socket.AF_INET):
                try:
                    r = requests.get(self.url4, timeout=self.timeout4,
                                     headers={'User-Agent': USER_AGENT})
                except requests.exceptions.RequestException as e:
                    self.log.error("Could not get IPv4 from %s: %s",
                                   self.url4, e)
                    got_ipv4 = False
                else:
                    try:
                        r.raise_for_status()
                    except requests.exceptions.HTTPError:
                        self.log.error("Received HTTP %d from %s: %s",
                                       r.status_code, self.url4, r.text)
                        got_ipv4 = False
                        err_ipv4 = True
                    else:
                        got_ipv4 = True
                        ipv4_text = r.text
            if got_ipv4:
                try:
                    ipv4 = ipaddress.IPv4Address(ipv4_text.strip())
                except ValueError:
                    self.log.error('Response from %s did not contain valid '
                                   'IPv4 address: "%s"', self.url4, ipv4_text)
                    got_ipv4 = False
                else:
                    self.notify_ipv4(ipv4)

        if self.want_ipv6():
            with RequestsFamilyRestriction(socket.AF_INET6):
                try:
                    r = requests.get(self.url6, timeout=self.timeout6,
                                     headers={'User-Agent': USER_AGENT})
                except requests.exceptions.RequestException as e:
                    self.log.error("Could not get IPv6 from %s: %s",
                                   self.url6, e)
                    got_ipv6 = False
                else:
                    try:
                        r.raise_for_status()
                    except requests.exceptions.HTTPError:
                        self.log.error("Received HTTP %d from %s: %s",
                                       r.status_code, self.url6, r.text)
                        got_ipv6 = False
                        err_ipv6 = True
                    else:
                        got_ipv6 = True
                        ipv6_text = r.text
            if got_ipv6:
                try:
                    ipv6 = ipaddress.IPv6Interface(
                        (ipv6_text.strip(), self.ipv6_prefix)).network
                except ValueError:
                    self.log.error('Response from %s did not contain valid '
                                   'IPv6 address: "%s"', self.url6, ipv6_text)
                    got_ipv6 = False
                else:
                    self.notify_ipv6(ipv6)

        # Raise for HTTP errors
        if err_ipv4:
            raise NotifyError(f"HTTP error for IPv4 in {self.name} notifier")
        if err_ipv6:
            raise NotifyError(f"HTTP error for IPv6 in {self.name} notifier")

        # Error if no wanted address was found
        if not (got_ipv4 or got_ipv6):
            raise NotifyError(f"Could not get any IP address for {self.name} "
                              "notifier")

        # Error for any missing wanted and needed addresses
        if self.need_ipv4() and not got_ipv4:
            raise NotifyError(f"Could not get IPv4 address for {self.name} "
                              "notifier")
        if self.need_ipv6() and not got_ipv6:
            raise NotifyError(f"Could not get IPv6 address for {self.name} "
                              "notifier")



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import ipaddress

import pytest

import doubles
from ruddr import PublishError


def test_init_hosts_str(empty_addrfile, data_dir):
    """Test init_hosts with raw string and the rest of the full flow"""
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        fetch_domain_ipv4s_result={
            'example.com': [(ipaddress.IPv4Address('1.2.3.4'), 1)],
            'foo.example.com': [(ipaddress.IPv4Address('1.2.3.4'), 2),
                                (ipaddress.IPv4Address('3.4.5.6'), 3)],
            'foo.bar.example.net': [(ipaddress.IPv4Address('3.4.5.6'), 4)],
        },
        fetch_all_ipv6s_result=[
            ('example.com', ipaddress.IPv6Address('1234::1'), 1),
            ('foo.example.com', ipaddress.IPv6Address('1234::2'), 2),
            ('foo.example.com', ipaddress.IPv6Address('3456::2'), 3),
            ('foo.bar.example.net', ipaddress.IPv6Address('3456::4'), 4),
            ('foo.bar.example.net', ipaddress.IPv6Address('3456::5'), 5),
            ('bar.example.org', ipaddress.IPv6Address('2345::6'), 6),
            ('bar.example.org', ipaddress.IPv6Address('4567::7'), 7),
        ],
        put_domain_ipv4_result={'example.com': None,
                                'foo.example.com': None,
                                'foo.bar.example.net': None},
        put_all_ipv6s_result=True,
    )
    updater.init_hosts('example.com foo.bar.example.net foo.example.com')

    updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))
    assert updater.fetch_all_ipv4s_count == 1
    assert updater.fetch_domain_ipv4s_calls == [
        'example.com',
        'foo.bar.example.net',
        'foo.example.com',
    ]
    assert updater.put_all_ipv4s_calls == []
    assert updater.put_domain_ipv4_calls == [
        ('example.com', ipaddress.IPv4Address('5.6.7.8'), 1),
        ('foo.bar.example.net', ipaddress.IPv4Address('5.6.7.8'), 4),
        ('foo.example.com', ipaddress.IPv4Address('5.6.7.8'), 2),
    ]

    updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))
    assert updater.fetch_all_ipv6s_count == 1
    assert updater.fetch_domain_ipv6s_calls == []
    assert updater.put_all_ipv6s_calls == [
        {'example.com': ([ipaddress.IPv6Address('5678::1')], 1),
         'foo.bar.example.net': ([ipaddress.IPv6Address('5678::4'),
                                  ipaddress.IPv6Address('5678::5')], 4),
         'foo.example.com': ([ipaddress.IPv6Address('5678::2')], 2),
         'bar.example.org': ([ipaddress.IPv6Address('2345::6'),
                              ipaddress.IPv6Address('4567::7')], 6)},
    ]
    assert updater.put_domain_ipv6s_calls == []


def test_init_hosts_list(empty_addrfile, data_dir):
    """Test init_hosts with list of domains and the rest of the full flow"""
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        fetch_all_ipv4s_result=[
            ('example.com', ipaddress.IPv4Address('1.2.3.4'), 1),
            ('foo.example.com', ipaddress.IPv4Address('1.2.3.4'), 2),
            ('foo.example.com', ipaddress.IPv4Address('3.4.5.6'), 3),
            ('foo.bar.example.net', ipaddress.IPv4Address('3.4.5.6'), 4),
            ('bar.example.org', ipaddress.IPv4Address('2.3.4.5'), 6),
            ('bar.example.org', ipaddress.IPv4Address('4.5.6.7'), 7),
        ],
        fetch_domain_ipv6s_result={
            'example.com': [(ipaddress.IPv6Address('1234::1'), 1)],
            'foo.example.com': [(ipaddress.IPv6Address('1234::2'), 2),
                                (ipaddress.IPv6Address('3456::2'), 3)],
            'foo.bar.example.net': [(ipaddress.IPv6Address('3456::4'), 4),
                                    (ipaddress.IPv6Address('3456::5'), 5)],
        },
        put_all_ipv4s_result=True,
        put_domain_ipv6s_result={'example.com': None,
                                 'foo.example.com': None,
                                 'foo.bar.example.net': None},
    )
    updater.init_hosts(
        ['example.com', 'foo.bar.example.net', 'foo.example.com']
    )

    updater.publish_ipv4(ipaddress.IPv4Address('5.6.7.8'))
    assert updater.fetch_all_ipv4s_count == 1
    assert updater.fetch_domain_ipv4s_calls == []
    assert updater.put_all_ipv4s_calls == [
        {'example.com': ([ipaddress.IPv4Address('5.6.7.8')], 1),
         'foo.bar.example.net': ([ipaddress.IPv4Address('5.6.7.8')], 4),
         'foo.example.com': ([ipaddress.IPv4Address('5.6.7.8')], 2),
         'bar.example.org': ([ipaddress.IPv4Address('2.3.4.5'),
                              ipaddress.IPv4Address('4.5.6.7')], 6)},
    ]
    assert updater.put_domain_ipv4_calls == []

    updater.publish_ipv6(ipaddress.IPv6Network('5678::/64'))
    assert updater.fetch_all_ipv6s_count == 1
    assert updater.fetch_domain_ipv6s_calls == [
        'example.com',
        'foo.bar.example.net',
        'foo.example.com',
    ]
    assert updater.put_all_ipv6s_calls == []
    assert updater.put_domain_ipv6s_calls == [
        ('example.com', [ipaddress.IPv6Address('5678::1')], 1),
        ('foo.bar.example.net', [ipaddress.IPv6Address('5678::4'),
                                 ipaddress.IPv6Address('5678::5')], 4),
        ('foo.example.com', [ipaddress.IPv6Address('5678::2')], 2),
    ]


def test_fetch_all_ipv4s(empty_addrfile, data_dir):
    """Test fetch_zone_ipv4s calls fetch_all_ipv4s"""
    result = [
        ('example.com', ipaddress.IPv4Address('1.2.3.4'), 1),
        ('foo.example.com', ipaddress.IPv4Address('1.2.3.4'), 2),
        ('foo.example.com', ipaddress.IPv4Address('3.4.5.6'), 3),
        ('foo.bar.example.net', ipaddress.IPv4Address('3.4.5.6'), 4),
        ('bar.example.org', ipaddress.IPv4Address('2.3.4.5'), 6),
        ('bar.example.org', ipaddress.IPv4Address('4.5.6.7'), 7),
    ]
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        fetch_all_ipv4s_result=result,
    )
    assert updater.fetch_zone_ipv4s('') == result
    assert updater.fetch_all_ipv4s_count == 1


def test_fetch_all_ipv4s_not_implemented(empty_addrfile, data_dir):
    """Test fetch_zone_ipv4s raises NotImplementedError when fetch_all_ipv4s
    is not implemented"""
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
    )
    with pytest.raises(NotImplementedError):
        updater.fetch_zone_ipv4s('')
    assert updater.fetch_all_ipv4s_count == 1


def test_fetch_all_ipv4s_publish_error(empty_addrfile, data_dir):
    """Test fetch_zone_ipv4s raises PublishError when fetch_all_ipv4s does"""
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        fetch_all_ipv4s_result=PublishError,
    )
    with pytest.raises(PublishError):
        updater.fetch_zone_ipv4s('')
    assert updater.fetch_all_ipv4s_count == 1


def test_fetch_domain_ipv4s(empty_addrfile, data_dir):
    """Test fetch_subdomain_ipv4s calls fetch_domain_ipv4s"""
    result = {
        'example.com': [(ipaddress.IPv4Address('1.2.3.4'), 1)],
    }
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        fetch_domain_ipv4s_result=result,
    )
    assert updater.fetch_subdomain_ipv4s(
        'example.com', ''
    ) == result['example.com']
    assert updater.fetch_domain_ipv4s_calls == ['example.com']


def test_fetch_domain_ipv4s_not_implemented(empty_addrfile, data_dir):
    """Test fetch_subdomain_ipv4s raises NotImplementedError when
    fetch_domain_ipv4s is not implemented"""
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
    )
    with pytest.raises(NotImplementedError):
        updater.fetch_subdomain_ipv4s('example.com', '')
    assert updater.fetch_domain_ipv4s_calls == ['example.com']


def test_fetch_domain_ipv4s_publish_error(empty_addrfile, data_dir):
    """Test fetch_subdomain_ipv4s raises PublishError when fetch_domain_ipv4s
    does"""
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        fetch_domain_ipv4s_result={'example.com': PublishError},
    )
    with pytest.raises(PublishError):
        updater.fetch_subdomain_ipv4s('example.com', '')
    assert updater.fetch_domain_ipv4s_calls == ['example.com']


def test_put_all_ipv4s(empty_addrfile, data_dir):
    """Test put_zone_ipv4s calls put_all_ipv4s"""
    arg = {'example.com': ([ipaddress.IPv4Address('5.6.7.8')], 1),
           'foo.bar.example.net': ([ipaddress.IPv4Address('5.6.7.8')], 4),
           'foo.example.com': ([ipaddress.IPv4Address('5.6.7.8')], 2),
           'bar.example.org': ([ipaddress.IPv4Address('2.3.4.5'),
                                ipaddress.IPv4Address('4.5.6.7')], 6)}
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        put_all_ipv4s_result=True,
    )
    updater.put_zone_ipv4s('', arg)
    assert updater.put_all_ipv4s_calls == [arg]


def test_put_all_ipv4s_not_implemented(empty_addrfile, data_dir):
    """Test put_zone_ipv4s raises NotImplementedError when put_all_ipv4s
    is not implemented"""
    arg = {'example.com': ([ipaddress.IPv4Address('5.6.7.8')], 1),
           'foo.bar.example.net': ([ipaddress.IPv4Address('5.6.7.8')], 4),
           'foo.example.com': ([ipaddress.IPv4Address('5.6.7.8')], 2),
           'bar.example.org': ([ipaddress.IPv4Address('2.3.4.5'),
                                ipaddress.IPv4Address('4.5.6.7')], 6)}
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
    )
    with pytest.raises(NotImplementedError):
        updater.put_zone_ipv4s('', arg)
    assert updater.put_all_ipv4s_calls == [arg]


def test_put_all_ipv4s_publish_error(empty_addrfile, data_dir):
    """Test put_zone_ipv4s raises PublishError when put_all_ipv4s does"""
    arg = {'example.com': ([ipaddress.IPv4Address('5.6.7.8')], 1),
           'foo.bar.example.net': ([ipaddress.IPv4Address('5.6.7.8')], 4),
           'foo.example.com': ([ipaddress.IPv4Address('5.6.7.8')], 2),
           'bar.example.org': ([ipaddress.IPv4Address('2.3.4.5'),
                                ipaddress.IPv4Address('4.5.6.7')], 6)}
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        put_all_ipv4s_result=PublishError,
    )
    with pytest.raises(PublishError):
        updater.put_zone_ipv4s('', arg)
    assert updater.put_all_ipv4s_calls == [arg]


def test_put_domain_ipv4(empty_addrfile, data_dir):
    """Test put_subdomain_ipv4 calls put_domain_ipv4"""
    address = ipaddress.IPv4Address('5.6.7.8')
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        put_domain_ipv4_result={'example.com': None},
    )
    updater.put_subdomain_ipv4('example.com', '', address, 2)
    assert updater.put_domain_ipv4_calls == [('example.com', address, 2)]


def test_put_domain_ipv4_not_implemented(empty_addrfile, data_dir):
    """Test put_subdomain_ipv4 raises NotImplementedError when put_domain_ipv4
    is not implemented"""
    address = ipaddress.IPv4Address('5.6.7.8')
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
    )
    with pytest.raises(NotImplementedError):
        updater.put_subdomain_ipv4('example.com', '', address, 2)
    assert updater.put_domain_ipv4_calls == [('example.com', address, 2)]


def test_put_domain_ipv4_publish_error(empty_addrfile, data_dir):
    """Test put_subdomain_ipv4 raises PublishError when put_domain_ipv4
    does"""
    address = ipaddress.IPv4Address('5.6.7.8')
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        put_domain_ipv4_result={'example.com': PublishError},
    )
    with pytest.raises(PublishError):
        updater.put_subdomain_ipv4('example.com', '', address, 2)
    assert updater.put_domain_ipv4_calls == [('example.com', address, 2)]


def test_fetch_all_ipv6s(empty_addrfile, data_dir):
    """Test fetch_zone_ipv6s calls fetch_all_ipv6s"""
    result = [
        ('example.com', ipaddress.IPv6Address('1234::1'), 1),
        ('foo.example.com', ipaddress.IPv6Address('1234::2'), 2),
        ('foo.example.com', ipaddress.IPv6Address('3456::2'), 3),
        ('foo.bar.example.net', ipaddress.IPv6Address('3456::4'), 4),
        ('foo.bar.example.net', ipaddress.IPv6Address('3456::5'), 5),
        ('bar.example.org', ipaddress.IPv6Address('2345::6'), 6),
        ('bar.example.org', ipaddress.IPv6Address('4567::7'), 7),
    ]
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        fetch_all_ipv6s_result=result,
    )
    assert updater.fetch_zone_ipv6s('') == result
    assert updater.fetch_all_ipv6s_count == 1


def test_fetch_all_ipv6s_not_implemented(empty_addrfile, data_dir):
    """Test fetch_zone_ipv6s raises NotImplementedError when fetch_all_ipv6s
    is not implemented"""
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
    )
    with pytest.raises(NotImplementedError):
        updater.fetch_zone_ipv6s('')
    assert updater.fetch_all_ipv6s_count == 1


def test_fetch_all_ipv6s_publish_error(empty_addrfile, data_dir):
    """Test fetch_zone_ipv6s raises PublishError when fetch_all_ipv6s does"""
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        fetch_all_ipv6s_result=PublishError,
    )
    with pytest.raises(PublishError):
        updater.fetch_zone_ipv6s('')
    assert updater.fetch_all_ipv6s_count == 1


def test_fetch_domain_ipv6s(empty_addrfile, data_dir):
    """Test fetch_subdomain_ipv6s calls fetch_domain_ipv6s"""
    result = {
        'example.com': [(ipaddress.IPv6Address('1234::1'), 1)],
    }
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        fetch_domain_ipv6s_result=result,
    )
    assert updater.fetch_subdomain_ipv6s(
        'example.com', ''
    ) == result['example.com']
    assert updater.fetch_domain_ipv6s_calls == ['example.com']


def test_fetch_domain_ipv6s_not_implemented(empty_addrfile, data_dir):
    """Test fetch_subdomain_ipv6s raises NotImplementedError when
    fetch_domain_ipv6s is not implemented"""
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
    )
    with pytest.raises(NotImplementedError):
        updater.fetch_subdomain_ipv6s('example.com', '')
    assert updater.fetch_domain_ipv6s_calls == ['example.com']


def test_fetch_domain_ipv6s_publish_error(empty_addrfile, data_dir):
    """Test fetch_subdomain_ipv6s raises PublishError when fetch_domain_ipv6s
    does"""
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        fetch_domain_ipv6s_result={'example.com': PublishError},
    )
    with pytest.raises(PublishError):
        updater.fetch_subdomain_ipv6s('example.com', '')
    assert updater.fetch_domain_ipv6s_calls == ['example.com']


def test_put_all_ipv6s(empty_addrfile, data_dir):
    """Test put_zone_ipv6s calls put_all_ipv6s"""
    arg = {'example.com': ([ipaddress.IPv6Address('5678::1')], 1),
           'foo.bar.example.net': ([ipaddress.IPv6Address('5678::4'),
                                    ipaddress.IPv6Address('5678::5')], 4),
           'foo.example.com': ([ipaddress.IPv6Address('5678::2')], 2),
           'bar.example.org': ([ipaddress.IPv6Address('2345::6'),
                                ipaddress.IPv6Address('4567::7')], 6)}
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        put_all_ipv6s_result=True,
    )
    updater.put_zone_ipv6s('', arg)
    assert updater.put_all_ipv6s_calls == [arg]


def test_put_all_ipv6s_not_implemented(empty_addrfile, data_dir):
    """Test put_zone_ipv6s raises NotImplementedError when put_all_ipv6s
    is not implemented"""
    arg = {'example.com': ([ipaddress.IPv6Address('5678::1')], 1),
           'foo.bar.example.net': ([ipaddress.IPv6Address('5678::4'),
                                    ipaddress.IPv6Address('5678::5')], 4),
           'foo.example.com': ([ipaddress.IPv6Address('5678::2')], 2),
           'bar.example.org': ([ipaddress.IPv6Address('2345::6'),
                                ipaddress.IPv6Address('4567::7')], 6)}
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
    )
    with pytest.raises(NotImplementedError):
        updater.put_zone_ipv6s('', arg)
    assert updater.put_all_ipv6s_calls == [arg]


def test_put_all_ipv6s_publish_error(empty_addrfile, data_dir):
    """Test put_zone_ipv6s raises PublishError when put_all_ipv6s does"""
    arg = {'example.com': ([ipaddress.IPv6Address('5678::1')], 1),
           'foo.bar.example.net': ([ipaddress.IPv6Address('5678::4'),
                                    ipaddress.IPv6Address('5678::5')], 4),
           'foo.example.com': ([ipaddress.IPv6Address('5678::2')], 2),
           'bar.example.org': ([ipaddress.IPv6Address('2345::6'),
                                ipaddress.IPv6Address('4567::7')], 6)}
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        put_all_ipv6s_result=PublishError,
    )
    with pytest.raises(PublishError):
        updater.put_zone_ipv6s('', arg)
    assert updater.put_all_ipv6s_calls == [arg]


def test_put_domain_ipv6s(empty_addrfile, data_dir):
    """Test put_subdomain_ipv6s calls put_domain_ipv6s"""
    addresses = [ipaddress.IPv6Address('5678::2')]
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        put_domain_ipv6s_result={'example.com': None},
    )
    updater.put_subdomain_ipv6s('example.com', '', addresses, 2)
    assert updater.put_domain_ipv6s_calls == [('example.com', addresses, 2)]


def test_put_domain_ipv6s_not_implemented(empty_addrfile, data_dir):
    """Test put_subdomain_ipv6s raises NotImplementedError when
    put_domain_ipv6s is not implemented"""
    addresses = [ipaddress.IPv6Address('5678::2')]
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
    )
    with pytest.raises(NotImplementedError):
        updater.put_subdomain_ipv6s('example.com', '', addresses, 2)
    assert updater.put_domain_ipv6s_calls == [('example.com', addresses, 2)]


def test_put_domain_ipv6s_publish_error(empty_addrfile, data_dir):
    """Test put_subdomain_ipv6s raises PublishError when put_domain_ipv6s
    does"""
    addresses = [ipaddress.IPv6Address('5678::2')]
    updater = doubles.MockTwoWayUpdater(
        'test_updater', empty_addrfile, data_dir,
        put_domain_ipv6s_result={'example.com': PublishError},
    )
    with pytest.raises(PublishError):
        updater.put_subdomain_ipv6s('example.com', '', addresses, 2)
    assert updater.put_domain_ipv6s_calls == [('example.com', addresses, 2)]



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Ruddr notifier that returns a statically-configured IP address"""

import ipaddress

from ruddr.exceptions import ConfigError
from .notifier import Notifier


class StaticNotifier(Notifier):
    """Ruddr notifier that returns a statically-configured IP address"""

    def __init__(self, name, config):
        super().__init__(name, config)

        # Static IPv4 address to use
        self.ipv4 = config.get('ipv4', None)
        if self.ipv4 is not None:
            try:
                self.ipv4 = ipaddress.IPv4Address(self.ipv4)
            except ValueError:
                self.log.critical("'ipv4' option contains an invalid IPv4 "
                                  "address")
                raise ConfigError(f"{self.name} notifier contains invalid "
                                  "address for 'ipv4' option") from None

        # Static IPv6 address with prefix to use. The prefix must be provided,
        # e.g. "1:2345::/64"
        self.ipv6 = config.get('ipv6', None)
        if self.ipv6 is not None:
            try:
                self.ipv6 = ipaddress.IPv6Network(self.ipv6)
            except ValueError:
                self.log.critical("'ipv6' option contains an invalid IPv6 "
                                  "address or no prefix")
                raise ConfigError(f"{self.name} notifier contains invalid "
                                  "address or no prefix for 'ipv6' option"
                                  ) from None

        if self.ipv4 is None and self.ipv6 is None:
            self.log.critical("No 'ipv4' or 'ipv6' option")
            raise ConfigError(f"{self.name} notifier requires either an IPv4 "
                              " or IPv6 address configured")

    def ipv4_ready(self):
        return self.ipv4 is not None

    def ipv6_ready(self):
        return self.ipv6 is not None

    def check_once(self):
        self.log.info("Checking IP addresses.")

        # No need to ensure the proper self.ipv4 or self.ipv6 variables are
        # set. The Notifier superclass already checked ipv4_ready() and
        # ipv6_ready().

        if self.want_ipv4():
            self.notify_ipv4(self.ipv4)
        if self.want_ipv6():
            self.notify_ipv6(self.ipv6)



#  Ruddr - Robotic Updater for Dynamic DNS Records
#  Copyright (C) 2023 Dominick C. Pastore
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""DDNS Manager: Initializes notifiers and updaters and manages the addrfile"""

import argparse
import importlib
import logging
import logging.handlers
import os.path
import signal
import sys
from typing import Optional, Any, Union, Dict, Tuple, cast

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

from . import Addrfile
from . import configuration
from .exceptions import NotifierSetupError, ConfigError, RuddrSetupError
from . import notifiers
from ruddr.util import sdnotify
from . import updaters


log = logging.getLogger('ruddr')


class DDNSManager:
    """Manages the rest of the Ruddr system. Creates notifiers and updaters and
    manages the addrfile.

    :param config: A :class:`~ruddr.Config` with the configuration to use

    :raises ConfigError: if configuration is not valid
    """

    def __init__(self, config: configuration.Config):
        try:
            config.finalize(validate_notifier_type, validate_updater_type)
        except ConfigError as e:
            log.critical("Config error: %s", e)
            raise
        self.config = config

        #: Addrfile manager
        addrfile_name = os.path.join(self.config.main['datadir'], 'addrfile')
        self.addrfile = Addrfile(addrfile_name)

        # Creates self.notifiers and self.updaters as dicts
        self.notifiers: Dict[
            str, notifiers.BaseNotifier
        ] = self._create_notifiers()
        self.updaters: Dict[
            str, updaters.BaseUpdater
        ] = self._create_updaters()

        self._discard_unused_notifiers()

    def _create_notifiers(self) -> Dict[str, notifiers.BaseNotifier]:
        """Initialize the notifiers. Assumes the notifiers have been previously
        imported by :func:`validate_notifier_type`."""
        notifiers_dict: Dict[str, notifiers.BaseNotifier] = dict()
        for name, config in self.config.notifiers.items():
            module = config.get('module')
            notifier_type = config['type']

            if module is None:
                notifier_class = notifiers.notifiers[notifier_type]
            else:
                notifier_class = notifiers.notifiers[(module, notifier_type)]

            notifier = notifier_class(name, config)
            notifiers_dict[name] = notifier
        return notifiers_dict

    def _create_updaters(self) -> Dict[str, updaters.BaseUpdater]:
        """Initialize the updaters. Assumes the updaters have been previously
        imported by :func:`validate_updater_type`."""
        updater_dict: Dict[str, updaters.BaseUpdater] = dict()
        for name, config in self.config.updaters.items():
            module = config.get('module')
            updater_type = config['type']

            if module is None:
                updater_class = updaters.updaters[updater_type]
            else:
                updater_class = updaters.updaters[(module, updater_type)]

            updater = updater_class(name, self.addrfile, config)
            updater.initial_update()
            self._attach_updater_notifier(updater, config)
            updater_dict[name] = updater
        return updater_dict

    def _attach_updater_notifier(self, updater, config):
        """Attach the given :class:`~ruddr.Updater` to its notifier(s).
        Assumes config is valid; that is, notifiers all exist.

        :param updater: The :class:`~ruddr.Updater` to be attached
        :param config: That :class:`~ruddr.Updater`'s config
        """
        ipv4_notifier_name = config.get('notifier4')
        ipv6_notifier_name = config.get('notifier6')

        if ipv4_notifier_name is not None:
            ipv4_notifier = self.notifiers[ipv4_notifier_name]
            ipv4_notifier.attach_ipv4_updater(updater.update_ipv4)
        if ipv6_notifier_name is not None:
            ipv6_notifier = self.notifiers[ipv6_notifier_name]
            ipv6_notifier.attach_ipv6_updater(updater.update_ipv6)

    def _discard_unused_notifiers(self):
        """Remove notifiers that are not attached to an updater"""
        for name in list(self.notifiers.keys()):
            if not (self.notifiers[name].want_ipv4() or
                    self.notifiers[name].want_ipv6()):
                log.warning("Notifier %s not attached to any updater", name)
                del self.notifiers[name]

    def start(self):
        """Start running all notifiers. Returns after they start. The notifiers
        will continue running in background threads.

        :raises NotifierSetupError: if a notifier fails to start
        """
        log.info("Starting all notifiers...")

        for name, notifier in self.notifiers.items():
            try:
                notifier.start()
            except NotifierSetupError:
                log.critical("Notifier %s failed to start. Stopping all "
                             "notifiers.", name)
                for running_notifier in self.notifiers.values():
                    running_notifier.stop()
                raise

        log.info("All notifiers started.")

    def do_notify(self):
        """Do an on-demand notify from all notifiers.

        Not all notifiers will support this, but most will.

        Does not raise any exceptions.
        """
        log.info("Checking once for all notifiers...")
        for notifier in self.notifiers.values():
            notifier.do_notify()
        log.info("Check for all notifiers complete.")

    def stop(self):
        """Stop all running notifiers gracefully. This will allow Python to
        exit naturally.

        Does not raise any exceptions, even if not yet started.
        """
        log.info("Stopping all notifiers...")
        for notifier in self.notifiers.values():
            notifier.stop()
        log.info("All notifiers stopped.")


def validate_notifier_type(module: Optional[str], type_: str) -> bool:
    """Check if a notifier type exists, importing it for :class:`DDNSManager`
    if it is not one of the built-in notifiers that comes with Ruddr

    :param module: ``None`` for built-in notifiers. Otherwise, the module the
                   notifier can be imported from.
    :param type_: The name of a built-in notifier or the class name of a
                  non-built-in notifier.
    :returns: ``True`` if the notifier exists, ``False`` otherwise
    """
    return _validate_updater_or_notifier_type(
        "notifier",
        notifiers.notifiers,
        module,
        type_
    )


def validate_updater_type(module: Optional[str], type_: str) -> bool:
    """Check if an updater type exists, importing it for :class:`DDNSManager`
    if it is not one of the built-in updaters that comes with Ruddr

    :param module: ``None`` for built-in updaters. Otherwise, the module the
                   updater can be imported from.
    :param type_: The name of a built-in updater or the class name of a
                  non-built-in updater.
    :returns: ``True`` if the updater exists, ``False`` otherwise
    """
    return _validate_updater_or_notifier_type(
        "updater",
        updaters.updaters,
        module,
        type_
    )


def _validate_updater_or_notifier_type(
    which: str,
    existing: Dict[Union[str, Tuple[str, str]], Any],
    module: Optional[str],
    type_: str,
) -> bool:
    """Check if an updater or notifier exists and import it

    :param which: ``"updater"`` or ``"notifier"``
    :param existing: The dict of already known updaters or notifiers and their
                     class. Keys are strings for built-in, (module, class) name
                     tuples for non-built-in
    :param module: ``None`` for built-in updaters or notifiers. Otherwise, the
                   module it can be imported from.
    :param type_: The name of a built-in updater or notifier or the class name
                  of a non-built-in one.
    :returns: ``True`` if the updater or notifier exists, ``False`` otherwise
    """
    if module is None:
        # Check if built-in or already imported with an entry point
        if type_ in existing:
            return True
        # Check if a ruddr entry point with this name exists
        discovered = entry_points(group=f"ruddr.{which}")
        try:
            entry_point = discovered[type_]
        except KeyError:
            return False
        existing[type_] = entry_point.load()
        return True

    # Check if already imported non-built-in
    if (module, type_) in existing:
        return True

    # Check if it's importable
    try:
        imported_module = importlib.import_module(module)
    except ImportError:
        return False
    try:
        imported_class = getattr(imported_module, type_)
    except AttributeError:
        return False
    existing[cast(Tuple[str, str], (module, type_))] = imported_class
    return True


def parse_args(argv):
    """Parse command line arguments

    :param argv: Either ``None`` or a list of arguments
    :returns: a :class:`argparse.Namespace` containing the parsed arguments
    """
    parser = argparse.ArgumentParser(
        description="Robotic Updater for Dynamic DNS Records",
        epilog="SIGUSR1 will cause a running instance to immediately check and"
               " update the current IP address(es) if possible",
    )
    parser.add_argument("-c", "--configfile", default="/etc/ruddr.conf",
                        help="Path to the config file")
    parser.add_argument("-d", "--debug-logs", action="store_true",
                        help="Increase verbosity of logging significantly")
    parser.add_argument("-s", "--stderr", action="store_true",
                        help="Log to stderr instead of syslog or file")
    return parser.parse_args(argv)


def main(argv=None):
    """Main entry point when run as a standalone program"""
    args = parse_args(argv)
    try:
        conf = configuration.read_config_from_path(args.configfile)
    except ConfigError as e:
        print("Config error:", e, file=sys.stderr)
        sys.exit(2)

    # Set up logging handler
    if args.stderr:
        logfile = 'stderr'
    else:
        logfile = conf.main.get('log', 'syslog')
    if logfile == 'syslog':
        log_handler = logging.handlers.SysLogHandler()
    elif logfile == 'stderr':
        log_handler = logging.StreamHandler()
    else:
        log_handler = logging.FileHandler(logfile)
    log.addHandler(log_handler)
    if args.debug_logs:
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)

    # Start up the actual DDNS code
    try:
        manager = DDNSManager(conf)
        manager.start()
    except RuddrSetupError:
        log.critical("Ruddr failed to start.")
        sys.exit(1)

    # Notify systemd, if applicable
    sdnotify.ready()

    # Do an immediate update on SIGUSR1
    def handle_sigusr1(sig, _):
        log.info("Received signal: %s", signal.Signals(sig).name)
        manager.do_notify()
    signal.signal(signal.SIGUSR1, handle_sigusr1)

    # Wait for SIGINT (^C) or SIGTERM
    def handle_signals(sig, _):
        log.info("Received signal: %s", signal.Signals(sig).name)
        sdnotify.stopping()
        manager.stop()
    signal.signal(signal.SIGINT, handle_signals)
    signal.signal(signal.SIGTERM, handle_signals)


if __name__ == '__main__':
    main()
